include lib.list
include lib.order

// Implementation of insertion sort.

val rec insert : ∀a, total_preorder⟨a⟩ ⇒ a ⇒ list⟨a⟩ ⇒ list⟨a⟩ =
  fun ord e l {
    case l {
      []   → e::[]
      x::r → if ord.cmp e x { e::l } else { x::(insert ord e r) }
    }
  }

val rec isort : ∀a, total_preorder⟨a⟩ ⇒ list⟨a⟩ ⇒ list⟨a⟩ =
  fun ord l {
    case l {
      []   → []
      x::l → insert ord x (isort ord l)
    }
  }

// Proof that insertion preserve sortedness.

val rec insert_lemma : ∀a, ∀ord∈total_preorder⟨a⟩,
    ∀e∈a, ∀l∈slist⟨a,ord⟩, sorted ord (insert ord e l) =
  fun ord e l {
    case l {
      []   → qed
      x::l →
        if ord.cmp e x {
          qed
        } else {
          use ord.total e x;
          case l {
            []   → qed
            y::r →
              use insert ord e r;
              if ord.cmp x y {
                deduce sorted ord l;
                use insert_lemma ord e l;
                if ord.cmp e y { qed } else { qed }
              } else { ✂ }
          }
        }
    }
  }

// Proof that insertion sort actually sorts.

val rec isort_theorem : ∀a, ∀ord∈total_preorder⟨a⟩,
    ∀l∈list⟨a⟩, sorted ord (isort ord l) =
  fun ord l {
    case l {
      []   → qed
      x::l → use isort_theorem ord l;
             use insert_lemma ord x (isort ord l)
    }
  }

// Insertion sort with a more precise type.

val isort : ∀a, ∀ord∈total_preorder⟨a⟩, list⟨a⟩ ⇒ slist⟨a,ord⟩ =
  fun ord l {
    use isort_theorem ord l;
    isort ord l
  }

// Now, [isort] has a more precise type than before.

check ∀a, ∀ord∈total_preorder⟨a⟩, list⟨a⟩ ⇒ slist⟨a,ord⟩
    ⊂ ∀a, total_preorder⟨a⟩ ⇒ list⟨a⟩ ⇒ list⟨a⟩
